"""
Taken from pandas.core.accessor.py, with minor adjustments.
"""


class CachedAccessor:
    """
    Custom property-like object.
    A descriptor for caching accessors.

    Parameters
    ----------
    name : str
        Namespace that will be accessed under, e.g. ``sc.plot``.
    accessor : cls
        Class with the extension methods.
    """

    def __init__(self, name: str, accessor) -> None:
        self._name = name
        self._accessor = accessor

    def __get__(self, obj, cls):
        if obj is None:
            # we're accessing the attribute of the class, i.e., Stairs.plot
            return self._accessor
        accessor_obj = self._accessor(obj)
        obj.__setattr__(self._name, accessor_obj)
        return accessor_obj



import operator
import warnings

import numpy as np
import pandas as pd

import staircase as sc
from staircase.core.ops import docstrings
from staircase.core.ops.common import _combine_stairs_via_values, requires_closed_match
from staircase.util import _sanitize_binary_operands
from staircase.util._decorators import Appender


@Appender(docstrings.negate_docstring, join="\n", indents=1)
def negate(self):
    data = -self._data if self._data is not None else None
    return sc.Stairs._new(
        initial_value=-self.initial_value,
        data=data,
        closed=self.closed,
    )


def _add_or_sub_deltas_no_mask(self, other, series_op, float_op):
    # assume self and other have ._data, and at least one has valid _deltas
    # if not other._valid_deltas:
    #     other._create_deltas()
    # if not self._valid_deltas:
    #     self._create_deltas()

    deltas = series_op(self._get_deltas(), other._get_deltas(), fill_value=0)

    new_instance = sc.Stairs._new(
        initial_value=float_op(self.initial_value, other.initial_value),
        data=pd.DataFrame({"delta": deltas}),
        closed=self.closed,
    )
    new_instance._remove_redundant_step_points()
    return new_instance


def _make_add_or_sub_func(docstring, series_op, float_op, series_rop):
    @Appender(docstring, join="\n", indents=1)
    @requires_closed_match
    def func(self, other):
        self, other = _sanitize_binary_operands(self, other)
        if self._data is None and other._data is None:
            return sc.Stairs._new(
                initial_value=float_op(self.initial_value, other.initial_value),
                data=None,
                closed=self.closed,
            )
        elif other._data is None:  # means self._data is not None
            if np.isnan(other.initial_value):
                data = None
            else:
                data = self._data.copy()
                if self._valid_values:
                    data["value"] = series_op(data["value"], other.initial_value)
            return sc.Stairs._new(
                initial_value=float_op(self.initial_value, other.initial_value),
                data=data,
                closed=self.closed,
            )
        elif self._data is None:  # means other._data is not None
            if np.isnan(self.initial_value):
                data = None
            else:
                data = other._data.copy()
                if other._valid_values:
                    data["value"] = series_rop(data["value"], self.initial_value)
                if other._valid_deltas:
                    data["delta"] = series_rop(data["delta"], 0)
            return sc.Stairs._new(
                initial_value=float_op(self.initial_value, other.initial_value),
                data=data,
                closed=other.closed,
            )
        # self._data or other._data exists
        elif self._has_na() or other._has_na():
            return _combine_stairs_via_values(self, other, series_op, float_op)
        elif self._valid_deltas or other._valid_deltas:
            return _add_or_sub_deltas_no_mask(self, other, series_op, float_op)
        elif self._valid_values and other._valid_values:
            return _combine_stairs_via_values(self, other, series_op, float_op)
        else:
            raise RuntimeError("This code should not execute")

    return func


add = _make_add_or_sub_func(
    docstrings.add_docstring,
    pd.Series.add,
    operator.add,
    pd.Series.radd,
)

subtract = _make_add_or_sub_func(
    docstrings.subtract_docstring,
    pd.Series.sub,
    operator.sub,
    pd.Series.rsub,
)


def _make_mul_div_func(docstring, series_op, float_op, series_rop, float_rop):
    @Appender(docstring, join="\n", indents=1)
    @requires_closed_match
    def func(self, other):
        def op_with_scalar(self, other, series_op, float_op):
            # other is scalar
            if other == 0 and series_op == pd.Series.divide:
                return sc.Stairs._new(np.nan, None, closed=self.closed)
            if self._data is None or np.isnan(other):
                data = None
            else:
                data = pd.DataFrame({"value": series_op(self._get_values(), other)})
                if series_op in (pd.Series.divide, pd.Series.rdiv):
                    data = data.replace(np.inf, np.nan)
            initial_value = float_op(self.initial_value, other)
            initial_value = initial_value if np.isfinite(initial_value) else np.nan
            return sc.Stairs._new(
                initial_value=initial_value,
                data=data,
                closed=self.closed,
            )

        self, other = _sanitize_binary_operands(self, other)
        if other._data is None:
            return op_with_scalar(self, other.initial_value, series_op, float_op)
        elif self._data is None:
            return op_with_scalar(other, self.initial_value, series_rop, float_rop)
        else:
            return _combine_stairs_via_values(self, other, series_op, float_op)

    return func


def float_rdiv(a, b):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        return np.divide(b, a)


multiply = _make_mul_div_func(
    docstrings.multiply_docstring,
    pd.Series.multiply,
    operator.mul,
    pd.Series.rmul,
    operator.mul,
)

divide = _make_mul_div_func(
    docstrings.divide_docstring,
    pd.Series.divide,
    np.divide,
    pd.Series.rdiv,
    float_rdiv,
)



import functools

import numpy as np
import pandas as pd

import staircase as sc
from staircase.core.exceptions import ClosedMismatchError


def _not_arithmetic_op(series_op):
    return series_op not in (
        pd.Series.add,
        pd.Series.sub,
        pd.Series.mul,
        pd.Series.div,
    )


def _reindex_deltas(new_index, deltas, initial_value):
    mask = deltas.where(np.isnan(deltas.values), 0)
    deltas_reindexed = deltas.reindex(new_index, fill_value=0)
    mask_initial_value = np.nan if np.isnan(initial_value) else 0
    mask_reindexed = _reindex_values(new_index, mask, mask_initial_value)
    return deltas_reindexed + mask_reindexed


def _reindex_values(new_index, values, initial_value):
    """
    Conform values to new index

    Parameters
    ----------
    new_index : pandas.Index
    values : pandas.Series
    initial_value : float

    Returns
    -------
    pandas.Series
    """
    first_step = values.index[0]
    new_values = values.reindex(new_index, method="ffill")
    new_values.loc[new_values.index < first_step] = initial_value
    return new_values


def _combine_step_series(
    series_1, series_2, initial_value_1, initial_value_2, series_op, kind
):
    # kind = "values" or "deltas"
    new_index = series_1.index.union(series_2.index)

    reindex_method = _reindex_values if kind == "values" else _reindex_deltas
    reindexed_series_1 = reindex_method(new_index, series_1, initial_value_1)
    reindexed_series_2 = reindex_method(new_index, series_2, initial_value_2)

    new_series = series_op(
        reindexed_series_1,
        reindexed_series_2,
    ).astype(float)

    if series_op == pd.Series.divide:
        new_series.replace(np.inf, np.nan, inplace=True)

    return new_series


def _combine_stairs_via_values(stairs1, stairs2, series_op, float_op):
    # self.values and other._values should be able to be created
    values = _combine_step_series(
        stairs1._get_values(),
        stairs2._get_values(),
        stairs1.initial_value,
        stairs2.initial_value,
        series_op,
        "values",
    )

    requires_manual_masking = _not_arithmetic_op(series_op)

    if requires_manual_masking and (stairs1._has_na() or stairs2._has_na()):
        mask = _combine_step_series(
            stairs1._get_values().isnull(),
            stairs2._get_values().isnull(),
            np.isnan(stairs1.initial_value),
            np.isnan(stairs2.initial_value),
            np.logical_or,
            "values",
        ).astype(bool)
        values.loc[mask] = np.nan

    if requires_manual_masking and (
        np.isnan(stairs1.initial_value) or np.isnan(stairs2.initial_value)
    ):
        initial_value = np.nan
    elif series_op == pd.Series.divide and stairs2.initial_value == 0:
        initial_value = np.nan
    else:
        initial_value = float_op(stairs1.initial_value, stairs2.initial_value) * 1

    if series_op == pd.Series.divide:
        values = values.replace(np.inf, np.nan).replace(-np.inf, np.nan)

    new_instance = sc.Stairs._new(
        initial_value=initial_value,
        data=pd.DataFrame({"value": values}),
        closed=stairs1.closed,
    )
    new_instance._remove_redundant_step_points()
    return new_instance


def _assert_closeds_equal(stairs1, stairs2):
    if (
        isinstance(stairs1, sc.Stairs)
        and isinstance(stairs2, sc.Stairs)
        and stairs1.number_of_steps != 0
        and stairs2.number_of_steps != 0
        and stairs1._closed != stairs2._closed
    ):
        raise ClosedMismatchError(stairs1, stairs2)


def requires_closed_match(func):
    @functools.wraps(func)
    def wrapper(stairs1, stairs2, *args, **kwargs):
        _assert_closeds_equal(stairs1, stairs2)
        return func(stairs1, stairs2, *args, **kwargs)

    return wrapper


def convert_string_args_to_timestamp(func):
    def convert_str(arg):
        if isinstance(arg, str):
            return pd.Timestamp(arg)
        return arg

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        args = [convert_str(arg) for arg in args]
        kwargs = {key: convert_str(val) for key, val in kwargs.items()}
        return func(self, *args, **kwargs)

    return wrapper



from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    AnyStr,
    Callable,
    Collection,
    Dict,
    Hashable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
)
from typing import Type as type_t
from typing import TypeVar, Union



from staircase.core.ops.arithmetic import add, divide, multiply, negate, subtract
from staircase.core.ops.logical import (
    invert,
    logical_and,
    logical_or,
    logical_xor,
    make_boolean,
)
from staircase.core.ops.masking import clip, fillna, isna, mask, notna, where
from staircase.core.ops.relational import eq, ge, gt, identical, le, lt, ne
from staircase.core.ops.rops import (
    logical_rand,
    logical_ror,
    logical_rxor,
    radd,
    rdivide,
    rmultiply,
    rsubtract,
)


def add_operations(cls):

    cls.negate = negate

    cls.add = add
    cls.subtract = subtract
    cls.clip = clip
    cls.mask = mask
    cls.where = where
    cls.isna = isna
    cls.notna = notna
    cls.fillna = fillna
    cls.multiply = multiply
    cls.divide = divide
    cls.eq = eq
    cls.ne = ne
    cls.lt = lt
    cls.gt = gt
    cls.le = le
    cls.ge = ge

    cls.__neg__ = negate
    cls.__add__ = add
    cls.__sub__ = subtract
    cls.__mul__ = multiply
    cls.__truediv__ = divide
    cls.__eq__ = eq
    cls.__ne__ = ne
    cls.__lt__ = lt
    cls.__gt__ = gt
    cls.__le__ = le
    cls.__ge__ = ge

    cls.invert = invert
    cls.logical_and = logical_and
    cls.logical_or = logical_or
    cls.logical_xor = logical_xor
    cls.make_boolean = make_boolean

    cls.__or__ = logical_or
    cls.__xor__ = logical_xor
    cls.__and__ = logical_and
    cls.__invert__ = invert

    cls.identical = identical

    cls.radd = radd
    cls.rdivide = rdivide
    cls.rmultiply = rmultiply
    cls.rsubtract = rsubtract
    cls.logical_rand = logical_rand
    cls.logical_ror = logical_ror
    cls.logical_rxor = logical_rxor

    cls.__radd__ = radd
    # cls.__rdiv__ = rdivide
    cls.__rtruediv__ = rdivide
    cls.__rmul__ = rmultiply
    cls.__rsub__ = rsubtract
    cls.__rand__ = logical_rand
    cls.__ror__ = logical_ror
    cls.__rxor__ = logical_rxor



import bisect
import operator

import numpy as np
import pandas as pd

import staircase as sc
from staircase.constants import inf
from staircase.core.ops import docstrings
from staircase.core.ops.common import (
    convert_string_args_to_timestamp,
    requires_closed_match,
)
from staircase.util import _replace_none_with_infs
from staircase.util._decorators import Appender


def _get_slice_index(self, lower, upper, lower_how, upper_how):
    # returns series
    if self._data is None:
        return -1, -1
    bisect_funcs = {
        "right": bisect.bisect_right,
        "left": bisect.bisect_left,
    }
    index_values = self._data.index.values
    lower_val_for_bisect = pd.Series([lower]).values[0]
    upper_val_for_bisect = pd.Series([upper]).values[0]
    left_index = bisect_funcs[lower_how](index_values, lower_val_for_bisect) - 1
    right_index = bisect_funcs[upper_how](index_values, upper_val_for_bisect)
    return left_index, right_index


@Appender(docstrings.clip_docstring, join="\n", indents=1)
@convert_string_args_to_timestamp
def clip(self, lower=-inf, upper=inf):
    lower, upper = _replace_none_with_infs((lower, upper))
    if not lower < upper:
        raise ValueError("'lower' must be strictly less than 'upper'.")
    if lower == -inf and upper == inf:
        return self
    left_index, right_index = _get_slice_index(
        self, lower, upper, lower_how="right", upper_how="left"
    )

    if right_index == -1:
        sliced_values = pd.Series(dtype="float64")
    else:
        sliced_values = self._get_values().iloc[max(0, left_index) : right_index]
    if upper != inf:
        sliced_values.loc[upper] = np.nan
    if lower != -inf and left_index < 0:
        sliced_values = pd.concat(
            [pd.Series([self.initial_value], index=[lower]), sliced_values]
        )
    elif sliced_values.index[0] < lower:
        index = pd.Series(sliced_values.index)
        index.iloc[0] = lower
        sliced_values.index = index

    data = pd.DataFrame({"value": sliced_values})

    initial_value = self.initial_value if lower == -inf else np.nan

    result = sc.Stairs._new(
        initial_value=initial_value,
        data=data,
        closed=self.closed,
    )
    result._remove_redundant_step_points()
    return result


def _maskify(self, inverse=False):
    func = np.array(0).__eq__ if not inverse else np.array(0).__ne__
    op = operator.ne if not inverse else operator.eq

    if self._data is None:
        data = None
    else:
        data = pd.DataFrame(
            {"value": self._get_values().where(func, np.nan).where(np.isnan, 0)}
        )

    return sc.Stairs._new(
        initial_value=np.nan if op(self.initial_value, 0) else 0,
        data=data,
        closed=self.closed,
    )


def _mask_stairs(self, other, inverse):
    def is_full_inverse_mask(initial_value):
        return initial_value == 0 or np.isnan(initial_value)

    full_mask_comparator = is_full_inverse_mask if inverse else float(0).__ne__
    if other._data is None:
        if full_mask_comparator(other.initial_value):
            return sc.Stairs(initial_value=np.nan)
        else:
            return self.copy()
    return _maskify(other, inverse=inverse) + self


def _make_mask_or_where_func(docstring, which):
    @convert_string_args_to_timestamp
    def handle_tuple_mask(self, left, right):
        return _mask_stairs(
            self, sc.Stairs().layer(start=left, end=right), inverse=False
        )

    @convert_string_args_to_timestamp
    def handle_tuple_where(self, left, right):
        left = -inf if left is None else left
        right = inf if right is None else right
        return clip(self, lower=left, upper=right)

    handle_tuple_func = {"mask": handle_tuple_mask, "where": handle_tuple_where}[which]
    inverse = which == "where"

    @Appender(docstring, join="\n", indents=1)
    def func(self, other):
        assert isinstance(other, tuple) or isinstance(
            other, sc.Stairs
        )  # TODO: exception
        if isinstance(other, tuple):
            if len(other) != 2:
                raise ValueError(
                    f"'other' tuple has length {len(other)} but requires length 2."
                )
            return handle_tuple_func(self, *other)
        return _mask_stairs(self, other, inverse=inverse)

    return func


mask = _make_mask_or_where_func(docstrings.mask_docstring, "mask")

where = _make_mask_or_where_func(docstrings.where_docstring, "where")


def _make_null_comparison_func(docstring, comp_func):
    @Appender(docstring, join="\n", indents=1)
    def func(self):
        initial_value = 1 if comp_func(self.initial_value) else 0
        if self._data is None:
            data = None
        else:
            data = pd.DataFrame(
                comp_func(self._get_values()) * 1, index=self._data.index
            )

        new_instance = sc.Stairs._new(
            initial_value=initial_value, data=data, closed=self.closed
        )
        new_instance._remove_redundant_step_points()
        return new_instance

    return func


# TODO: test
isna = _make_null_comparison_func(docstrings.isna_docstring, np.isnan)

# TODO: test
notna = _make_null_comparison_func(docstrings.notna_docstring, lambda x: ~np.isnan(x))


def _make_data_fillna_method(self, value):
    # value is a string
    initial_value = self.initial_value
    if self._data is None:
        data = None
    else:
        values = self._get_values().copy()
        fillmethod = {
            "pad": pd.Series.ffill,
            "ffill": pd.Series.ffill,
            "backfill": pd.Series.bfill,
            "bfill": pd.Series.bfill,
        }[value]
        if value in ("pad", "ffill") and np.isnan(values.iloc[0]):
            values.iloc[0] = self.initial_value
        values = fillmethod(values)
        if value in ("backfill", "bfill") and np.isnan(self.initial_value):
            initial_value = values.iloc[0]
        data = pd.DataFrame({"value": values})
    return initial_value, data


def _make_data_fillna_scalar(self, value):
    if np.isnan(self.initial_value):
        initial_value = value
    else:
        initial_value = self.initial_value

    if self._data is None:
        data = None
    else:
        values = self._get_values().copy().fillna(value=value)
        data = pd.DataFrame({"value": values})
    return initial_value, data


@requires_closed_match
def _fillna_with_stairs(self, value):
    # value is Stairs
    return self.fillna(0) + value * self.isna()


# TODO: test
@Appender(docstrings.fillna_docstring, join="\n", indents=1)
def fillna(self, value):

    if isinstance(value, sc.Stairs):
        return _fillna_with_stairs(self, value)

    if isinstance(value, str):
        initial_value, data = _make_data_fillna_method(self, value)
    else:
        initial_value, data = _make_data_fillna_scalar(self, value)

    new_instance = sc.Stairs._new(
        initial_value=initial_value, data=data, closed=self.closed
    )
    new_instance._remove_redundant_step_points()
    return new_instance



from staircase.core.accessor import CachedAccessor
from staircase.core.stats.statistic import _max as max
from staircase.core.stats.statistic import _min as min
from staircase.core.stats.statistic import (
    agg,
    corr,
    cov,
    integral,
    mean,
    median,
    mode,
    std,
    value_sums,
    values_in_range,
    var,
)


def add_methods(cls):
    from staircase.core.stats.distribution import Dist

    cls.value_sums = value_sums
    cls.values_in_range = values_in_range
    cls.cov = cov
    cls.corr = corr
    cls.agg = agg

    cls.integral = integral
    cls.mean = mean
    cls.median = median
    cls.mode = mode
    cls.std = std
    cls.var = var

    cls.dist = CachedAccessor("dist", Dist)



import datetime

import numpy as np
import pandas as pd

import staircase as sc
from staircase.constants import inf


def _is_datetime_like(x):
    return isinstance(
        x,
        (
            pd.Timestamp,
            pd.Timedelta,
            datetime.timedelta,
            datetime.datetime,
            datetime.date,
            np.datetime64,
            np.timedelta64,
        ),
    )


def _sanitize_binary_operands(self, other, copy_other=False):
    if not isinstance(self, sc.Stairs):
        self = sc.Stairs(initial_value=self, closed=other.closed)
    else:
        self = self.copy()

    if not isinstance(other, sc.Stairs):
        other = sc.Stairs(initial_value=other, closed=self.closed)
    else:
        if copy_other:
            other = other.copy()
    return self, other


def _get_lims(stairs, closed):

    if stairs._closed == "left":
        return {
            "both": ("right", "right"),
            "left": ("right", "left"),
            "right": ("right", "right"),
            "neither": ("right", "left"),
        }[closed]
    else:
        return {
            "both": ("left", "left"),
            "left": ("left", "left"),
            "right": ("right", "left"),
            "neither": ("right", "left"),
        }[closed]


def _replace_none_with_infs(tuple_):
    if tuple_ == (-inf, inf):
        return tuple_
    left, right = tuple_
    if left is None:
        left = -inf
    if right is None:
        right = inf
    return (left, right)



_symbol_map = {
    "addition": "+",
    "subtraction": "-",
    "multiplication": r"*",
    "division": "/",
    "lt": "<",
    "le": "<=",
    "gt": ">",
    "ge": ">=",
    "eq": "==",
    "ne": "!=",
    "and": "&",
    "or": "|",
    "xor": "^",
    "negate": "-",
    "invert": "~",
    "raddition": "+",
    "rsubtraction": "-",
    "rmultiplication": r"*",
    "rdivision": "/",
    "rand": "&",
    "ror": "|",
    "rxor": "^",
}

_see_also_map = {
    "addition": "Stairs.subtract, Stairs.radd",
    "subtraction": "Stairs.add, Stairs.rsubtract",
    "multiplication": "Stairs.divide, Stairs.rmultiply",
    "division": "Stairs.multiply, Stairs.rdivide",
    "lt": "Stairs.gt, Stairs.le, Stairs.ge",
    "le": "Stairs.lt, Stairs.gt, Stairs.ge",
    "gt": "Stairs.lt, Stairs.le, Stairs.ge",
    "ge": "Stairs.lt, Stairs.gt, Stairs.le",
    "eq": "Stairs.ne, Stairs.identical",
    "ne": "Stairs.eq, Stairs.identical",
    "and": "Stairs.logical_or, Stairs.logical_rand",
    "or": "Stairs.logical_and, Stairs.logical_xor, Stairs.logical_ror",
    "xor": "Stairs.logical_or, Stairs.logical_rxor",
    "negate": "Stairs.subtract",
    "make_boolean": "Stairs.invert",
    "invert": "Stairs.make_boolean",
    "raddition": "Stairs.rsubtract, Stairs.add",
    "rsubtraction": "Stairs.radd, Stairs.subtract",
    "rmultiplication": "Stairs.rdivide, Stairs.multiply",
    "rdivision": "Stairs.rmultiply, Stairs.divide",
    "rand": "Stairs.logical_ror, Stairs.logical_and",
    "ror": "Stairs.logical_rand, Stairs.logical_or",
    "rxor": "Stairs.logical_ror, Stairs.logical_xor",
}

_relationship_map = {
    "lt": "strictly less than",
    "le": "less than, or equal to,",
    "gt": "strictly greater than",
    "ge": "greater than, or equal to,",
    "eq": "equal to",
    "ne": "not equal to",
}

_plot_titles_map = {
    "addition": ["s1", "s2", "s1+s2"],
    "subtraction": ["s1", "s2", "s1-s2"],
    "multiplication": ["s1", "s2", "s1*s2"],
    "division": ["s1", "s2", "s1/(s2+2)"],
    "lt": ["s1", "s2", "s1<s2"],
    "le": ["s1", "s2", "s1<=s2"],
    "gt": ["s1", "s2", "s1>s2"],
    "ge": ["s1", "s2", "s1>=s2"],
    "eq": ["s1", "s2", "s1==s2"],
    "ne": ["s1", "s2", "s1!=s2"],
    "and": ["s1", "s2", "s1&s2"],
    "or": ["s1", "s2", "s1|s2"],
    "xor": ["s1", "s2", "s1^s2"],
    "invert": ["s2", "~s2"],
    "make_boolean": ["s2", "s2.make_boolean()"],
    "negate": ["s1", "-s1"],
    "raddition": ["s1", "2+s1"],
    "rsubtraction": ["s1", "2-s1"],
    "rmultiplication": ["s1", "2*s1"],
    "rdivision": ["s1", "1/(s1+2)"],
    "rand": ["s1", "s2", "s2&s1"],
    "ror": ["s1", "s2", "s2|s1"],
    "rxor": ["s1", "s2", "s2^s1"],
}

_example = """
.. plot::
    :context: close-figs
    :include-source: False

    {setup}
    >>> stair_list = [{plots}]
    >>> fig, axes = plt.subplots(nrows=1, ncols={ncols},  figsize=({width},3), sharey=True, sharex=True, tight_layout=True, dpi=400)
    >>> for ax, title, stair_instance in zip(axes, {plot_titles}, stair_list):
    ...     stair_instance.plot(ax=ax, arrows=True)
    ...     ax.set_title(title)
"""


def _gen_example(plot_titles, setup=""):
    # plot_titles is a list of strings
    ncols = len(plot_titles)
    plots = ", ".join(plot_titles)
    return _example.format(
        plots=plots,
        ncols=ncols,
        plot_titles=plot_titles,
        setup=setup,
        width=ncols + 5,
    )


_arithmetic_binop_docstring_header = """
A binary operator facilitating the {operation} of step functions (and floats).
"""

_logical_binop_docstring_header = """
Returns a boolean-valued step function indicating where *self* {operation} *other* are non-zero.
"""

_relational_binop_docstring_header = """
Returns a boolean-valued step function indicating where *self* is {relationship} *other*.
"""

_negate_header = """
An operator which produces a new Stairs instance representing the multiplication of the step function by -1.
"""

_invert_header = """
Returns a boolean-valued step function indicating where *self* is zero-valued.
"""

_common_docstring = """
Equivalent to {op_string}.  See examples below.
{params}
Returns
-------
:class:`Stairs`
     new instance representing the result of {op_string}

See Also
--------
{{see_also}}

Examples
--------
{{example}}
"""

params = """
Parameters
----------
other :  int, float, or :class:`Stairs`
"""
_common_unop_docstring = _common_docstring.format(
    op_string=r"{symbol}\ *self*", params="\n"
)
_common_binop_docstring = _common_docstring.format(
    op_string="*self* {symbol} *other*", params="\n".join([" ", params, " "])
)
_common_rbinop_docstring = _common_docstring.format(
    op_string="*other* {symbol} *self*", params="\n".join([" ", params, " "])
)

_arithmetic_binops = ("addition", "subtraction", "multiplication", "division")
_arithmetic_rbinops = ("raddition", "rsubtraction", "rmultiplication", "rdivision")
_relational_binops = ("lt", "le", "gt", "ge", "eq", "ne")
_logical_binops = {"and", "or", "xor"}
_logical_rbinops = {"rand", "ror", "rxor"}


def _get_header(operation):
    if operation in _arithmetic_binops:
        result = _arithmetic_binop_docstring_header.format(operation=operation)
        return result
    if operation in _arithmetic_rbinops:
        result = _arithmetic_binop_docstring_header.format(operation=operation[1:])
        return result
    if operation in _relational_binops:
        relationship = _relationship_map[operation]
        return _relational_binop_docstring_header.format(relationship=relationship)
    if operation in _logical_binops:
        return _logical_binop_docstring_header.format(operation=operation)
    if operation in _logical_rbinops:
        return _logical_binop_docstring_header.format(operation=operation[1:])
    if operation == "negate":
        return _negate_header
    if operation == "invert":
        return _invert_header


def _gen_op_docstring(operation):
    symbol = _symbol_map[operation]
    see_also = _see_also_map[operation]
    example = _gen_example(_plot_titles_map[operation])

    header = _get_header(operation)
    if operation in ("negate", "invert"):
        body = _common_unop_docstring
    elif operation in (
        "raddition",
        "rsubtraction",
        "rmultiplication",
        "rdivision",
        "rand",
        "ror",
        "rxor",
    ):
        body = _common_rbinop_docstring
    else:
        body = _common_binop_docstring
    docstring = header + body
    return docstring.format(
        symbol=symbol,
        see_also=see_also,
        example=example,
    )


add_docstring = _gen_op_docstring("addition")
subtract_docstring = _gen_op_docstring("subtraction")
multiply_docstring = _gen_op_docstring("multiplication")
divide_docstring = _gen_op_docstring("division")
negate_docstring = _gen_op_docstring("negate")
lt_docstring = _gen_op_docstring("lt")
le_docstring = _gen_op_docstring("le")
gt_docstring = _gen_op_docstring("gt")
ge_docstring = _gen_op_docstring("ge")
eq_docstring = _gen_op_docstring("eq")
ne_docstring = _gen_op_docstring("ne")
logical_and_docstring = _gen_op_docstring("and")
logical_or_docstring = _gen_op_docstring("or")
logical_xor_docstring = _gen_op_docstring("xor")
invert_docstring = _gen_op_docstring("invert")
radd_docstring = _gen_op_docstring("raddition")
rsubtract_docstring = _gen_op_docstring("rsubtraction")
rmultiply_docstring = _gen_op_docstring("rmultiplication")
rdivide_docstring = _gen_op_docstring("rdivision")
logical_rand_docstring = _gen_op_docstring("rand")
logical_ror_docstring = _gen_op_docstring("ror")
logical_rxor_docstring = _gen_op_docstring("rxor")


identical_docstring = """
Returns True if *self* and *other* represent the same step functions.

Returns
-------
boolean

See Also
--------
Stairs.eq, Stairs.ne

Examples
--------

>>> s1.identical(s1)
True
>>> s1.identical(s1.copy())
True
>>> s1.identical(s1.copy().layer(1,2))
False
"""

_make_boolean_docstring_body = """
Returns a boolean-valued step function indicating where *self* is non-zero.

Returns
-------
:class:`Stairs`
    A new instance representing where *self* is non-zero

See Also
--------
Stairs.invert

Examples
--------
"""

make_boolean_docstring = _make_boolean_docstring_body + _gen_example(
    ["s2", "s2.make_boolean()"]
)


clip_docstring = """
Returns a copy of *self* which is undefined everywhere outside of [lower, upper]

Parameters
----------
lower : int, float, pandas.Timedelta, pandas.Timestamp or str
    lower bound of the interval.  A string argument is assumed to be a string representation
    of a pandas.Timestamp
upper : int, float, pandas.Timedelta, pandas.Timestamp or str
    upper bound of the interval.  A string argument is assumed to be a string representation
    of a pandas.Timestamp

Returns
-------
:class:`Stairs`
    Returns a copy of *self* which is undefined everywhere outside of [lower, upper)

Examples
--------

{plot}
>>> s1.clip(2,4).mean()
0.5
""".format(
    plot=_gen_example(["s1", "s1.clip(2,4)"])
)

# _mask_where_2_col_example = """
# .. plot::
#     :context: close-figs

#     >>> masker = {masker}
#     >>> stairs_list = [s2, s2.{func}(masker)]
#     >>> fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(6,3), sharey=True, sharex=True, tight_layout=True, dpi=400)
#     >>> for ax, title, stair_instance in zip(axes, ("s2", "s2.{func}(masker)"), stairs_list):
#     ...     stair_instance.plot(ax=ax)
#     ...     ax.set_title(title)
# """

# _mask_where_3_col_example = """
# .. plot::
#     :context: close-figs

#     >>> masker = {masker}
#     >>> stairs_list = [s2, masker, s2.{func}(masker)]
#     >>> fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(8,3), sharey=True, sharex=True, tight_layout=True, dpi=400)
#     >>> for ax, title, stair_instance in zip(axes, ("s2", "masker", "s2.{func}(masker)"), stairs_list):
#     ...     stair_instance.plot(ax=ax)
#     ...     ax.set_title(title)
# """

# mask_examples = f"""

# {_mask_where_2_col_example.format(masker="(2,4)", func="mask")}

# {_mask_where_3_col_example.format(masker="sc.Stairs(initial_value=2).layer(1,2,-1).layer(4,5,-2)", func="where")}
# """

# where_examples = f"""

# {_mask_where_2_col_example.format(masker="(1,4)", func="where")}

# {_mask_where_3_col_example.format(masker="sc.Stairs().layer(1,2).layer(4,5,-2)", func="where")}
# """

mask_examples = "\n".join(
    [
        _gen_example(["s2", "s2.mask((2,4))"], ">>> masker = (2,4)"),
        _gen_example(
            ["s2", "masker", "s2.mask(masker)"],
            ">>> masker = sc.Stairs(initial_value=2).layer(1,2,-2).layer(4,5,-2)",
        ),
    ]
)

where_examples = "\n".join(
    [
        _gen_example(["s2", "s2.where((1,4))"], ">>> masker = (1,4)"),
        _gen_example(
            ["s2", "masker", "s2.where(masker)"],
            ">>> masker = sc.Stairs().layer(1,2).layer(4,5,-2)",
        ),
    ]
)

_mask_where_docstring = """
Returns a new step function where *self* has been {operation} by *other*.

Note when *other* is as Stairs instance, is is considered equivalent to its
boolean value, that is, the result of :meth:`Stairs.make_boolean`.  As a result any part
of the domain where *other* is {value}, or not defined, will be undefined in the resulting
step function.

Also note that ``.{func}((a,b))`` is equivalent to ``.{func}(sc.Stairs().layer(a,b))`` but defers
to a faster implementation.

Parameters
----------
other : :class:`Stairs`, or tuple
    If *other* is a tuple *(a,b)* then it is assumed that a < b and that
    *a* and *b* both belong to the domain of the step function represented by *self*.
    An exception is when *a* or *b* are strings, in which case they are assumed to
    be pandas.Timestamp and the above conditions apply.

Returns
-------
:class:`Stairs`

See Also
--------
{see_also}

Examples
--------
{examples}
"""

mask_docstring = _mask_where_docstring.format(
    operation="masked",
    value="zero-valued",
    func="mask",
    examples=mask_examples,
    see_also="Stairs.where",
)

where_docstring = _mask_where_docstring.format(
    operation="'inversely masked'",
    value="not zero-valued",
    func="where",
    examples=where_examples,
    see_also="Stairs.mask, Stairs.clip",
)


_isna_notna_docstring = """
Returns a new 'boolean valued' step function indicating where the
domain of *self* is not defined.

The value of the step function is {def_value} where *self* is defined, and {not_def_value} where
*self* is not defined.

Returns
-------
:class:`Stairs`

See Also
--------
{see_also}

Examples
--------
{examples}
"""

isna_docstring = _isna_notna_docstring.format(
    def_value=0,
    not_def_value=1,
    see_also="Stairs.notna",
    examples=_gen_example(["s3", "s3.isna()"]),
)
notna_docstring = _isna_notna_docstring.format(
    def_value=1,
    not_def_value=0,
    see_also="Stairs.isna",
    examples=_gen_example(["s3", "s3.notna()"]),
)

fillna_examples = "\n".join(
    [
        _gen_example(["s3", "s3.fillna(0.5)"]),
        _gen_example(["s3", 's3.fillna("ffill")']),
        _gen_example(["s3", 's3.fillna("bfill")']),
        _gen_example(["s3", "s2", "s3.fillna(s2)"]),
    ]
)

fillna_docstring = """
Define values for (a copy of) *self* where it is undefined.

Parameters
----------
value : {{int, float, "backfill", "bfill", "pad", "ffill", :class:`Stairs`}}
    If *value* is and int, float or :class:`Stairs`, then it is used to provide values
    where *self* is undefined.  If *value* is a string then it indicates
    a method for propagating values of the step function across undefined
    intervals:

    - ``pad / ffill`` propagate last defined value forward
    - ``backfill / bfill`` propagate next defined value backward

Returns
-------
:class:`Stairs`

See Also
--------
Stairs.isna

Examples
--------
{examples}
""".format(
    examples=fillna_examples
)



"""

staircase
==============

staircase is a MIT licensed library, written in pure-Python, for
modelling step functions. See :ref:`Getting Started <getting_started>` for more information.
"""

from __future__ import annotations

import warnings
from typing import Any, Callable

import numpy as np
import pandas as pd
from pandas.api.types import (
    is_datetime64_dtype,
    is_number,
    is_numeric_dtype,
    is_timedelta64_dtype,
)
from typing_extensions import Literal

from staircase import docstrings
from staircase.constants import inf
from staircase.core import stats
from staircase.core.accessor import CachedAccessor
from staircase.core.layering import _check_args_dtypes
from staircase.plotting.accessor import PlotAccessor
from staircase.util import _replace_none_with_infs
from staircase.util._decorators import Appender


def _make_deltas_from_vals(init_val, vals: pd.DataFrame) -> pd.Series:
    if not np.isnan(vals).any():
        result = pd.Series(
            np.diff((np.append([init_val], vals.values))), dtype="float64"
        )
    else:
        temp = pd.Series(np.append([init_val], vals.values), dtype="float64")
        result = pd.Series.add(
            temp[pd.notnull(temp)].diff(), temp[pd.isnull(temp)], fill_value=0
        )[1:]
    if np.isnan(init_val):
        result.iloc[0] = vals.iloc[0]
    result.index = vals.index
    return result


def _make_vals_from_deltas(init_val: float, deltas: pd.DataFrame):
    base = 0 if np.isnan(init_val) else init_val
    return deltas.cumsum() + base


class Stairs:

    class_name = "Stairs"

    @Appender(docstrings.Stairs_docstring, join="\n", indents=2)
    def __init__(
        self,
        frame: pd.DataFrame | None = None,
        start: int | None = None,
        end=None,
        value=None,
        initial_value=0,
        closed: Literal["left", "right"] = "left",
    ):
        assert frame is None or isinstance(frame, pd.DataFrame)
        self._data = None
        self._valid_deltas = False
        self._valid_values = False
        self._masked = False
        self._closed = closed
        self.initial_value = initial_value
        self._clear_cache()

        if any([x is not None for x in (start, end, value)]):
            self.layer(start, end, value, frame)

    def _clear_cache(self):
        self.dist._reset()
        self._integral_and_mean = None

    @classmethod
    def _new(
        cls,
        initial_value: float,
        data: pd.DataFrame,
        closed: Literal["left", "right"] = "left",
    ) -> Stairs:
        new_instance = cls(closed=closed)
        new_instance.initial_value = initial_value
        new_instance._data = data
        new_instance._valid_deltas = False if data is None else "delta" in data.columns
        new_instance._valid_values = False if data is None else "value" in data.columns
        return new_instance

    @classmethod
    def from_values(
        cls,
        initial_value: float = 0,
        values: pd.Series | None = None,
        closed: Literal["left", "right"] = "left",
    ) -> Stairs:
        """
        Construct :class:`Stairs` from :class:`pandas.Series`.

        Parameters
        ----------
        initial_value : float, default 0
            The value of the step function at negative infinity.
        values : :class:`pandas.Series`
            The step function values' when approaching the change points from the right
        closed : {"left", "right"}
            Indicates whether the half-open intervals comprising the step function should be interpreted
            as left-closed or right-closed.

        Returns
        -------
        :class:`Stairs`
        """

        if not isinstance(values, pd.Series) or values.empty:
            raise ValueError("values must be a not empty Series")

        _check_args_dtypes(values.index)

        if is_numeric_dtype(values.index) and np.isinf(values.index).any():
            raise ValueError("Invalid value for Series index")

        if not is_numeric_dtype(values) or not is_number(initial_value):
            raise ValueError("Invalid dtype for from_values()")

        if not values.index.is_monotonic_increasing:
            raise ValueError("Series index must be monotonic")

        series_values_inf_mask = np.isinf(values)
        if series_values_inf_mask.any():
            values = values.replace([np.inf], np.nan)
            warnings.warn("Infinity values detected and have been converted to NaN")

        new_instance = cls(closed=closed)
        new_instance.initial_value = initial_value
        new_instance._data = values.to_frame("value")
        new_instance._valid_deltas = False
        new_instance._valid_values = True
        return new_instance

    def _has_na(self) -> bool | np.array:
        return np.isnan(self._data.values).any() or np.isnan(self.initial_value)

    def _create_values(self) -> Stairs:
        assert self._valid_deltas
        self._data["value"] = _make_vals_from_deltas(
            self.initial_value, self._data["delta"]
        )
        self._valid_values = True
        return self

    def _create_deltas(self) -> Stairs:
        assert self._valid_values
        self._data["delta"] = _make_deltas_from_vals(
            self.initial_value, self._data["value"]
        )
        self._valid_deltas = True
        return self

    def _get_deltas(self) -> pd.Series:
        if self._data is None:
            return pd.Series(dtype="float64")
        if not self._valid_deltas:
            self._create_deltas()
        return self._data["delta"]

    def _get_values(self) -> pd.Series:
        if self._data is None:
            return pd.Series(dtype="float64")
        if not self._valid_values:
            self._create_values()
        return self._data["value"]

    @property
    def closed(self):
        return self._closed

    @property
    @Appender(stats.docstrings.ecdf_example, join="\n", indents=2)
    def ecdf(self):
        """
        Calculates an `empirical cumulative distribution function <https://en.wikipedia.org/wiki/Empirical_distribution_function>`_
        for the corresponding step function values.

        Returns
        -------
        :class:`ECDF`

        See Also
        --------
        Stairs.hist
        Stairs.percentile
        Stairs.fractile
        """
        return self.dist.ecdf

    @property
    def fractile(self):
        """
        This property returns a :class:`staircase.Fractiles` object.

        Fractiles is a subclass of Stairs.  Although it inherits all methods of Stairs,
        it is expected that it will typically only be used as a callable class, to
        calculate the x fractile with the following parameters.

        Parameters
        ----------
        x : float or array-like of float
            Percentile or sequence of fractiles to compute,.  Must be between 0 and 1.

        Returns
        -------
        float or :class:`numpy.ndarray` of floats

        See Also
        --------
        Stairs.percentile, Stairs.quantiles

        Examples
        --------

        .. plot::
            :context: close-figs

            >>> sf = sc.Stairs().layer([0,1,2,3,4], [6,7,8,9,10])
            >>> fig, axes = plt.subplots(ncols=2, figsize=(7,3), sharey=True)
            >>> sf.plot(axes[0])
            >>> axes[0].set_title("sf")
            >>> sf.fractile.plot(axes[1])
            >>> axes[1].set_title("fractile function")

        >>> sf.fractile
        <staircase.Fractiles, id=2151414503360>

        >>> sf.fractile(0.5)
        3

        >>> sf.percentile([0.25, 0.5, 0.75])
        array([2., 3., 4.])
        """
        return self.dist.fractile

    @property
    def percentile(self):
        """
        This property returns a :class:`staircase.Percentiles` object.

        Percentiles is a subclass of Stairs.  Although it inherits all methods of Stairs,
        it is expected that it will typically only be used as a callable class, to
        calculate the x-th percentile with the following parameters.

        Parameters
        ----------
        x : int or array-like of int
            Percentile or sequence of percentiles to compute,.  Must be between 0 and 100.

        Returns
        -------
        float or :class:`numpy.ndarray` of floats

        See Also
        --------
        Stairs.fractile, Stairs.quantiles

        Examples
        --------

        .. plot::
            :context: close-figs

            >>> sf = sc.Stairs().layer([0,1,2,3,4], [6,7,8,9,10])
            >>> fig, axes = plt.subplots(ncols=2, figsize=(7,3), sharey=True)
            >>> sf.plot(axes[0])
            >>> axes[0].set_title("sf")
            >>> sf.percentile.plot(axes[1])
            >>> axes[1].set_title("percentile function")

        >>> sf.percentile
        <staircase.Percentiles, id=2151414790944>

        >>> sf.percentile(50)
        3

        >>> sf.percentile([25,50,75])
        array([2., 3., 4.])
        """
        return self.dist.percentile

    @Appender(stats.docstrings.hist_example, join="\n", indents=2)
    def hist(self, bins="unit", closed="left", stat="sum"):
        """
        Calculates histogram data for the corresponding step function values

        Parameters
        ----------
        bins : "unit", sequence or :class:`pandas.IntervalIndex`
            If *bins* is "unit" then the histogram bins will have unit length and cover the range
            of step function values.  If *bins* is a sequence, it defines a monotonically
            increasing array of bin edges.  If *bins* are defined by :class:`pandas.IntervalIndex`
            they should be non-overlapping and monotonic increasing.
        closed : {"left", "right"}, default "left"
            Indicates whether the histogram bins are left-closed right-open
            or right-closed left-open. Only relevant when *bins* is not a :class:`pandas.IntervalIndex`
        stat : {"sum", "frequency", "density", "probability"}, default "sum"
            The aggregate statistic to compute in each bin.  Inspired by :meth:`seaborn.histplot` stat parameter.
                - ``sum`` the magnitude of observations
                - ``frequency`` values of the histogram are divided by the corresponding bin width
                - ``density`` normalises values of the histogram so that the area is 1
                - ``probability`` normalises values so that the histogram values sum to 1

        Returns
        -------
        :class:`pandas.DataFrame`
        """
        return self.dist.hist(bins=bins, closed=closed, stat=stat)

    def quantiles(self, q: int):
        """
        Returns an array of q-quantiles.

        Quantiles are cut points which divide a distribution into continuous intervals with equal probabilities.
        The 2-quantile is more commonly known as the median.
        The 4-quantiles are more commonly known as quartiles.
        The 100-quantiles are more commonly known as percentiles.

        Parameters
        ----------
        q : int

        Returns
        -------
        :class:`numpy.ndarray` of floats

        See Also
        --------
        Stairs.percentile, Stairs.fractile

        Examples
        --------

        .. plot::
            :context: close-figs

            >>> sf = sc.Stairs().layer([0,1,2,3,4], [6,7,8,9,10])
            >>> fig, axes = plt.subplots(ncols=2, figsize=(7,3))
            >>> sf.plot(axes[0])
            >>> axes[0].set_title("sf")
            >>> sf.fractile.plot(axes[1])
            >>> axes[1].set_title("fractile function")

        >>> sf.quantiles(4)
        array([2., 3., 4.])
        """
        return self.dist.quantiles(q)

    @Appender(stats.docstrings.simple_max_example, join="\n", indents=2)
    def max(self) -> float:
        """
        The maximum of the step function.

        Returns
        -------
        float
            The maximum of the step function

        See Also
        --------
        Stairs.min, Stairs.values_in_range
        """
        return stats.max(self)

    @Appender(stats.docstrings.simple_min_example, join="\n", indents=2)
    def min(self) -> float:
        """
        The minimum of the step function.

        Returns
        -------
        float
            The minimum of the step function

        See Also
        --------
        Stairs.max, Stairs.values_in_range
        """
        return stats.min(self)

    plot = CachedAccessor("plot", PlotAccessor)

    @property
    def step_changes(self) -> pd.Series:
        """
        A pandas Series of key, value pairs of indicating where step changes occur in the step function, and the change in value

        Returns
        -------
        :class:`pandas.Series`

        See Also
        --------
        Stairs.step_points
        Stairs.step_values
        Stairs.number_of_steps

        Examples
        --------

        .. plot::
            :context: close-figs

            >>> s1.plot()
            >>> s1.step_changes
            1    1
            2   -1
            3    1
            4   -2
            5    1
            dtype: int64
        """
        return self._get_deltas().copy()

    @property
    def step_values(self):
        """
        A pandas Series of key, value pairs of indicating where step changes occur in the step function, and
        the limit of the step function when it approaches these points from the right.

        Returns
        -------
        :class:`pandas.Series`

        See Also
        --------
        Stairs.step_points
        Stairs.step_changes
        Stairs.number_of_steps

        Examples
        --------

        .. plot::
            :context: close-figs

            >>> s1.plot()
            >>> s1.step_values
            1    1
            2    0
            3    1
            4   -1
            5    0
            dtype: int64
        """
        return self._get_values().copy()

    @property
    def step_points(self) -> np.array:
        """
        A numpy arrauy of domain values indicating where step changes occur in the step function.

        Returns
        -------
        :class:`numpy.ndarray`

        See Also
        --------
        Stairs.step_values
        Stairs.step_changes
        Stairs.number_of_steps

        Examples
        --------

        .. plot::
            :context: close-figs

            >>> s1.plot()
            >>> s1.step_values
            array([1, 2, 3, 4, 5], dtype=int64)
        """
        if self._data is None:
            return np.array([])
        return self._data.index.values

    @Appender(docstrings.examples.number_of_steps_example, join="\n", indents=2)
    @property
    def number_of_steps(self) -> int:
        """
        Calculates the number of step changes

        Returns
        -------
        int

        See Also
        --------
        Stairs.step_changes
        Stairs.step_values
        Stairs.step_points
        """
        return len(self.step_points)

    def _remove_redundant_step_points(self) -> Stairs:

        # preferred over values method
        def remove_via_deltas() -> None:
            remove_index = (
                self._data["delta"].isna() & self._data["delta"].isna().shift()
            ) | (self._data["delta"] == 0)
            self._data = self._data.loc[~remove_index]

        # do we just make deltas and then run above method?
        def remove_via_values() -> None:
            remove_index = (
                self._data["value"].isna() & self._data["value"].isna().shift()
            ) | (self._data["value"] == self._data["value"].shift())
            if (self._data["value"].iloc[0] == self.initial_value) or (
                pd.isnull(self._data.loc[:, "value"].iloc[0])
                and pd.isnull(self.initial_value)
            ):
                remove_index.iloc[0] = True
            self._data = self._data.loc[~remove_index]

        if self._data is not None:
            if self._valid_deltas:
                remove_via_deltas()
            elif self._valid_values:
                remove_via_values()
            else:
                assert False, "no deltas or values valid!"
            if len(self._data) == 0:
                self._data = None

        return self

    def copy(self) -> Stairs:
        """
        Returns a deep copy of this Stairs instance

        Returns
        -------
        :class:`Stairs`
        """
        new_instance = Stairs._new(
            initial_value=self.initial_value,
            data=self._data.copy() if self._data is not None else None,
            closed=self.closed,
        )
        return new_instance

    def __bool__(self) -> bool:
        """
        Return True if and only if step function has a value of 1 everywhere.

        Returns
        -------
        boolean
        """
        if self.initial_value == 1 and self.number_of_steps == 0:
            return True
        return False

    @Appender(docstrings.examples.describe_example, join="\n", indents=2)
    def describe(
        self,
        where: tuple[float | int, float | int]
        | list[float | int, float | int] = (
            -inf,
            inf,
        ),
        percentiles: list[float | int] | tuple[float | int] = (25, 50, 75),
    ) -> pd.Series:
        """
        Generate descriptive statistics for the step function values over a specified domain.

        Parameters
        ----------
        where : tuple or list of length two, optional
            Indicates the domain interval over which to evaluate the step function.
            Default is (-sc.inf, sc.inf) or equivalently (None, None).
        percentiles: array-like of float, default [25, 50, 70]
            The percentiles to include in output.  Numbers should be in the range 0 to 100.

        Returns
        -------
        :class:`pandas.Series`

        See Also
        --------
        Stairs.mean, Stairs.std, Stairs.min, Stairs.percentile, Stairs.max
        """
        where = _replace_none_with_infs(where)
        stairs = self if where == (-inf, inf) else self.clip(*where)

        return pd.Series(
            {
                **{
                    "unique": stairs.percentile.clip(0, 100).number_of_steps - 1,
                    "mean": stairs.mean,
                    "std": stairs.std,
                    "min": stairs.min,
                },
                **{f"{perc}%": stairs.percentile(perc) for perc in percentiles},
                **{"max": stairs.max},
            }
        )

    @Appender(docstrings.examples.shift_example, join="\n", indents=2)
    def shift(self, delta: pd.Series | int | float) -> Stairs:
        """
        Returns a stairs instance corresponding to a horizontal translation by delta

        If delta is positive the corresponding step function is moved right.
        If delta is negative the corresponding step function is moved left.

        Parameters
        ----------
        delta : int, float or pandas.Timedelta
            the amount by which to translate.  A pandas.Timedelta is only valid when using dates.
            If using dates and delta is an int or float, then it is interpreted as a number of hours.

        Returns
        -------
        :class:`Stairs`

        See Also
        --------
        Stairs.diff
        """
        if self._data is None:
            return Stairs(initial_value=self.initial_value)
        return Stairs._new(
            initial_value=self.initial_value,
            data=self._data.set_index(self._data.index + delta),
            closed=self.closed,
        )

    @Appender(docstrings.examples.diff_example, join="\n", indents=2)
    def diff(self, delta: int | float | pd.Timedelta) -> Stairs:
        """
        Returns a stairs instance corresponding to the difference between the step function corresponding to *self*
        and the same step-function translated by delta.

        Parameters
        ----------
        delta : int, float or pandas.Timedelta
            the amount by which to translate.  A pandas.Timestamp is only valid when using dates.
            If using dates and delta is an int or float, then it is interpreted as a number of hours.

        Returns
        -------
        :class:`Stairs`

        See Also
        --------
        Stairs.shift
        """
        return self - self.shift(delta)

    @Appender(docstrings.examples.rolling_mean_example, join="\n", indents=2)
    def rolling_mean(
        self,
        window: tuple[int | int] = (0, 0),
        where: tuple[float | float] = (-inf, inf),
    ) -> pd.Series:
        """
        Returns coordinates defining rolling mean

        The rolling mean of a step function is a continous piece-wise linear function, hence it can
        be described by a sequence of x,y coordinates which mark where function changes gradient.  These
        x,y coordinates are returned as a :class:`pandas.Series` which could then be used with
        :meth:`matplotlib.axes.Axes.plot`, or equivalent, to visualise.

        A rolling mean requires a window around a point x (referred to as the focal point) to be defined.
        In this implementation the window is defined by two values paired into an array-like parameter called *window*.
        These two numbers are the distance from the focal point to the left boundary of the window, and the right boundary
        of the window respectively.  This allows for trailing windows, leading windows and everything between
        (including a centred window).

        Parameters
        ----------
        window : array-like of int, float or pandas.Timedelta
            should be length of 2. Defines distances from focal point to window boundaries.
        where : tuple or list of length two, optional
            Indicates the domain interval over which to evaluate the step function.
            Default is (-sc.inf, sc.inf) or equivalently (None, None).

        Returns
        -------
        :class:`pandas.Series`

        See Also
        --------
        Stairs.mean
        """
        where = _replace_none_with_infs(where)
        assert len(window) == 2, "Window should be a listlike object of length 2."
        left_delta, right_delta = window
        lower, upper = where
        clipped = self.clip(lower, upper)
        if clipped._data is None:
            return pd.Series([clipped.initial_value] * 2, index=where)
        step_points = clipped._data.index
        sample_points = pd.Index.union(
            step_points - left_delta,
            step_points - right_delta,
        )
        ii = pd.IntervalIndex.from_arrays(
            sample_points + left_delta, sample_points + right_delta
        )
        s = pd.Series(
            clipped.slice(ii).mean().values,
            index=sample_points,
        )
        if lower != -inf:
            s = s.loc[s.index >= lower - left_delta]
        if upper != inf:
            s = s.loc[s.index <= upper - right_delta]
        return s

    def to_frame(self) -> pd.DataFrame:
        """
        Returns a pandas.DataFrame with columns 'start', 'end' and 'value'

        The rows of the dataframe can be interpreted as the interval definitions
        which make up the step function.

        Returns
        -------
        :class:`pandas.DataFrame`

        Examples
        --------

        .. plot::
            :context: close-figs

            >>> s1.plot()

        >>> s1.to_frame()
        0  -inf    1      0
        1     1    2      1
        2     2    3      0
        3     3    4      1
        4     4    5     -1
        5     5  inf      0
        """
        if self._data is None:
            starts = [-inf]
            ends = [inf]
            values = [self.initial_value]
        else:
            if not self._valid_values:
                self._create_values()
            step_points = self._data.index
            starts = [-inf] + step_points.to_list()
            ends = step_points.to_list() + [inf]
            values = np.append(self.initial_value, self._data["value"].values)
        return pd.DataFrame({"start": starts, "end": ends, "value": values})

    def pipe(self, func: Callable, *args, **kwargs) -> Any:
        """
        Applies func and returns the result.

        Primarily intended to facilitate method chaining.

        Parameters
        ----------
        func : callable
            Function to apply to *self*.
        args : , optional
            Positional arguments passed into *func*.
        kwargs : mapping, optional
            A dictionary of keyword arguments passed into *func*.

        Returns
        -------
        object
            return type of *func*
        """
        return func(self, *args, **kwargs)

    def __str__(self) -> str:
        """
        Return str(self)
        """
        return f"<staircase.{self.class_name}, id={id(self)}>"

    def __repr__(self) -> str:
        """
        Return string representation of Stairs
        """
        return str(self)

    def __call__(self, *args, **kwargs):
        return self.sample(*args, **kwargs)


def _add_operations() -> None:
    from staircase.core import layering, ops, sampling, slicing

    ops.add_operations(Stairs)
    stats.add_methods(Stairs)
    sampling.add_methods(Stairs)
    layering.add_methods(Stairs)
    slicing.add_methods(Stairs)



from __future__ import annotations

from typing import Callable

import numpy as np
import pandas as pd
from pandas.api.types import is_list_like
from typing_extensions import Literal

import staircase as sc
from staircase.core.ops.masking import clip
from staircase.core.stats.statistic import _get_stairs_method
from staircase.docstrings import slicing as docstrings
from staircase.util._decorators import Appender


class StairsSlicer:
    def __init__(self, stairs: sc.Stairs, interval_index):
        self._stairs = stairs
        self._interval_index = interval_index
        self._slices: pd.Series | None = None

    def _create_slices(self) -> None:
        slices = self._interval_index.map(lambda i: clip(self._stairs, i.left, i.right))
        self._slices = pd.Series(slices, index=self._interval_index)

    def _ensure_slices(self) -> StairsSlicer:
        if self._slices is None:
            self._create_slices()
        return self

    @Appender(docstrings.agg_docstring, join="\n", indents=1)
    def agg(self, funcs) -> pd.DataFrame:
        if isinstance(funcs, str):
            funcs = [funcs]
        self._ensure_slices()
        df = pd.DataFrame(index=self._slices.index)
        for func in funcs:
            df[func] = getattr(self, func)()
        return df

    @Appender(docstrings.apply_docstring, join="\n", indents=1)
    def apply(self, func: Callable, *args, **kwargs) -> pd.Series:
        self._ensure_slices()
        return self._slices.apply(func, args=args, **kwargs)

    @Appender(docstrings._docstrings["max"], join="\n", indents=1)
    def max(self) -> float:
        result = self._max()
        if self._stairs._closed == "left" and self._interval_index.closed in (
            "right",
            "both",
        ):
            result = np.maximum(result, self._stairs(self._interval_index.right))
        elif self._stairs._closed == "right" and self._interval_index.closed in (
            "left",
            "both",
        ):
            result = np.maximum(result, self._stairs(self._interval_index.left))
        return result

    @Appender(docstrings._docstrings["min"], join="\n", indents=1)
    def min(self) -> float:
        result = self._min()
        if self._stairs._closed == "left" and self._interval_index.closed in (
            "right",
            "both",
        ):
            result = np.minimum(result, self._stairs(self._interval_index.right))
        elif self._stairs._closed == "right" and self._interval_index.closed in (
            "left",
            "both",
        ):
            result = np.minimum(result, self._stairs(self._interval_index.left))
        return result

    @Appender(docstrings.hist_docstring, join="\n", indents=1)
    def hist(self, *args, **kwargs):
        self._ensure_slices()
        zero = (
            self._stairs._data.index[0] - self._stairs._data.index[0]
        )  # hack to get 0 or pd.Timedelta(0)
        return self._slices.apply(sc.Stairs.hist, *args, **kwargs).fillna(zero)

    @Appender(docstrings.resample_docstring, join="\n", indents=1)
    def resample(self, func: str):
        if not self._interval_index.is_non_overlapping_monotonic:
            raise ValueError(
                "Slices must be monotonic increasing (ascending order) and not overlapping"
            )
        self._ensure_slices()
        new_values = getattr(self, func)()
        left_bound = self._slices.index.left.min()
        right_bound = self._slices.index.right.max()
        stairs_na = self._stairs.isna().mask((left_bound, right_bound)).fillna(0)
        return (
            self._stairs.mask((left_bound, right_bound))
            .fillna(0)
            .mask(stairs_na)
            .layer(new_values.index.left, new_values.index.right, new_values.values)
        )


def make_slice_method(method_name: str) -> Callable:

    func = _get_stairs_method(method_name)
    docstring = docstrings._docstrings.get(method_name, "")

    @Appender(docstring, join="\n", indents=1)
    def method(self):
        self._ensure_slices()
        return self._slices.map(func)

    return method


for method_name in ["_max", "_min", "mean", "median", "integral", "mode"]:
    method = make_slice_method(method_name)
    setattr(StairsSlicer, method_name, method)


def slice(
    self: sc.Stairs, cuts, closed: Literal["left", "right", "both", "neither"] = "left"
) -> StairsSlicer:
    """
    Slice the step function into pieces.

    A slice corresponding to an interval *(a,b)* will be equal to *self* everywhere
    in *(a,b)* and undefined elsewhere.

    Parameters
    ----------
    self : Stairs : class
    cuts : sequence, :class:`pandas.IntervalIndex`, :class:`pandas.PeriodIndex`.
        Used to slice the step function.  If *cuts* is a sequence then it should comprised
        of monotonically increasing values from the step function domain.
    closed : {'left', 'right', 'both', 'neither'}
        Only relevant if *cuts* is not a :class:`pandas.IntervalIndex`, and indicates
        if the intervals derived from *cuts* should closed on the left-side, right-side, both or neither.

    Returns
    -------
    :class:`StairsSlicer`

    Examples
    --------

    .. plot::
        :context: close-figs

        >>> s3.plot(arrows=True)

    >>> s3.slice(np.linspace(1,5,9))
    <staircase.core.slicing.StairsSlicer at 0x20053f07400>

    >>> s3.slice(np.linspace(1,5,9)).mean()
    [1.0, 1.5)    1.0
    [1.5, 2.0)    1.0
    [2.0, 2.5)    0.0
    [2.5, 3.0)    NaN
    [3.0, 3.5)    NaN
    [3.5, 4.0)    1.0
    [4.0, 4.5)   -1.0
    [4.5, 5.0)   -1.0
    dtype: float64

    >>> s3.slice(np.linspace(0,6,4)).agg(["min", "max"])
                min  max
    [0.0, 2.0)  0.0  1.0
    [2.0, 4.0)  0.0  1.0
    [4.0, 6.0) -1.0  0.0

    >>> s3.slice(np.linspace(0,6,4)).hist(bins=[-1, -0.5, 0, 0.5, 1, 1.5])
                [-1.0, -0.5)  [-0.5, 0.0)  [0.0, 0.5)  [0.5, 1.0)  [1.0, 1.5)
    [0.0, 2.0)           0.0          0.0         1.0         0.0         1.0
    [2.0, 4.0)           0.0          0.0         0.5         0.0         0.5
    [4.0, 6.0)           1.0          0.0         1.0         0.0         0.0
    """
    if isinstance(cuts, pd.IntervalIndex):
        ii = cuts
    elif isinstance(cuts, pd.PeriodIndex):
        end_times = cuts.end_time + pd.Timedelta(
            "1ns"
        )  # PeriodIndex leaves a 1ns gap between intervals
        ii = pd.IntervalIndex.from_arrays(cuts.start_time, end_times, closed=closed)
    elif is_list_like(cuts):
        ii = pd.IntervalIndex.from_breaks(cuts, closed=closed)

    return StairsSlicer(self, ii)


def add_methods(cls):
    cls.slice = slice



from __future__ import annotations


class NegInf:
    def __init__(self):
        pass

    def __lt__(self, other) -> bool:
        assert not isinstance(other, NegInf)
        return True

    def __gt__(self, other) -> bool:
        assert not isinstance(other, (Inf, NegInf))
        return False

    def __neg__(self) -> Inf:
        return inf

    def __str__(self) -> str:
        return "-inf"


class Inf:
    def __init__(self):
        pass

    def __gt__(self, other) -> bool:
        assert not isinstance(other, Inf)
        return True

    def __lt__(self, other) -> bool:
        assert not isinstance(other, (Inf, NegInf))
        return False

    def __neg__(self) -> NegInf:
        return neginf

    def __str__(self) -> str:
        return "inf"


neginf = NegInf()
inf = Inf()



import warnings

import numpy as np
import pandas as pd
from pandas.api.types import is_list_like

import staircase as sc
from staircase.constants import inf
from staircase.core.ops.masking import _get_slice_index
from staircase.core.stats import docstrings
from staircase.docstrings import examples
from staircase.util import _get_lims, _replace_none_with_infs
from staircase.util._decorators import Appender


def _cache_integral_and_mean(self):
    if self._data is None or len(self._data) < 2:
        self._integral_and_mean = np.nan, np.nan
    else:
        value_sums = self.value_sums(group=False)
        value_sums = value_sums[value_sums.index.notnull()]
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                # integral calculated this way to force an overflow error if required
                # pandas.TimedeltaIndex * int is not raising an overflow error when it ideally should
                integral = np.array(
                    [x * y for x, y in zip(value_sums.index, value_sums)]
                ).sum()
                mean = integral / value_sums.sum()
                self._integral_and_mean = (integral, mean)
        except (OverflowError, ValueError) as exc:
            integral = None
            mean = ((value_sums / value_sums.sum()) * value_sums.index).sum()
            self._integral_and_mean = (integral, mean)
            raise exc


@Appender(docstrings.integral_docstring, join="\n", indents=1)
def integral(self):
    if self._integral_and_mean is None:
        try:
            _cache_integral_and_mean(self)
        except (OverflowError, ValueError):
            raise OverflowError(
                "Integral calculation results in overflow error.  Consider scaling down step function values to accommodate."
            )
    return self._integral_and_mean[0]


@Appender(docstrings.mean_docstring, join="\n", indents=1)
def mean(self):
    if self._integral_and_mean is None:
        try:
            _cache_integral_and_mean(self)
        except OverflowError:
            pass
    return self._integral_and_mean[1]


@Appender(docstrings.median_docstring, join="\n", indents=1)
def median(self):
    return self.dist.percentile(50)


@Appender(docstrings.value_sums_docstring, join="\n", indents=1)
def value_sums(self, dropna=True, group=True):

    if self._data is None:
        return None

    value_sums = pd.Series(
        np.diff(self._data.index.values), index=self._get_values().iloc[:-1]
    )
    # .values used to avoid a strange numpy Future Warning
    if group:
        result = value_sums.groupby(value_sums.index.values).sum()
    else:
        return value_sums
    if not dropna:
        isna = value_sums.index.isna()
        if isna.any():
            na_val = value_sums.where(isna).first_valid_index()
            result[na_val] = value_sums[isna].sum()
    return result


@Appender(docstrings.mode_docstring, join="\n", indents=1)
def mode(self):
    return value_sums(self).idxmax()


@Appender(docstrings.var_docstring, join="\n", indents=1)
def var(self):
    percentile_minus_mean = self.dist.percentile - mean(self)
    values = percentile_minus_mean._get_values()
    squared_values = values * values
    return (
        sc.Stairs._new(
            initial_value=0,
            data=pd.DataFrame({"value": squared_values}),
        ).agg("integral", (0, 100))
        / 100
    )


@Appender(docstrings.std_docstring, join="\n", indents=1)
def std(self):
    return np.sqrt(var(self))


@Appender(docstrings.values_in_range_docstring, join="\n", indents=1)
def values_in_range(self, where=(-inf, inf), closed=None):
    where = _replace_none_with_infs(where)
    assert len(where) == 2, "Parameter 'where' should be list or tuple of length 2."
    if closed is None:
        closed = self._closed
    lower, upper = where
    lower_how, upper_how = _get_lims(self, closed)
    left_index, right_index = _get_slice_index(self, lower, upper, lower_how, upper_how)
    if right_index == -1:
        return np.array([self.initial_value])
    values = self._get_values().iloc[max(0, left_index) : right_index]
    if left_index < 0 and not np.isnan(self.initial_value):
        values = np.append([self.initial_value], values)
    unique = np.unique(values)
    return unique[~np.isnan(unique)]


def _min(
    self,
    where=(-inf, inf),
    closed=None,
):
    where = _replace_none_with_infs(where)
    if closed is None:
        closed = self._closed
    return min(self.values_in_range(where, closed))


def _max(
    self,
    where=(-inf, inf),
    closed=None,
):
    where = _replace_none_with_infs(where)
    if closed is None:
        closed = self._closed
    return max(self.values_in_range(where, closed))


@Appender(docstrings.agg_docstring, join="\n", indents=1)
def agg(self, name, where=(-inf, inf), closed=None):

    where = _replace_none_with_infs(where)
    stairs = self if where == (-inf, inf) else self.clip(*where)

    def apply(func):
        if isinstance(func, str):
            name = func
            func = _get_stairs_method(name)
        else:
            name = func.__name__
        if name in ("min", "max"):
            return name, func(self, where=where, closed=closed)
        else:
            return name, func(stairs)

    if is_list_like(name):
        return pd.Series({func: calc for func, calc in map(apply, name)})
    return apply(name)[1]


@Appender(examples.cov_example, join="\n", indents=1)
def cov(self, other, where=(-inf, inf), lag=0, clip="pre"):
    """
    Calculates either covariance, autocovariance or cross-covariance.

    The calculation is between two step functions described by *self* and *other*.
    If lag is None or 0 then covariance is calculated, otherwise cross-covariance is calculated.
    Autocovariance is a special case of cross-covariance when *other* is equal to *self*.

    Parameters
    ----------
    other: :class:`Stairs`
        the stairs instance with which to compute the covariance
    where : tuple or list of length two, optional
    Indicates the domain interval over which to perform the calculation.
        Default is (-sc.inf, sc.inf) or equivalently (None, None).
    lag : int, float, pandas.Timedelta
        A pandas.Timedelta is only valid when domain is date-like.
    clip : {'pre', 'post'}, default 'pre'
        Only relevant when lag is non-zero.  Determines if the domain is applied before or
        after *other* is translated.  If 'pre' then the domain over which the calculation
        is performed is the overlap of the original domain and the translated domain.

    Returns
    -------
    float
        The covariance (or cross-covariance) between *self* and *other*

    See Also
    --------
    Stairs.corr, staircase.cov, staircase.corr
    """
    where = list(_replace_none_with_infs(where))
    if lag != 0:
        assert clip in ["pre", "post"]
        if clip == "pre" and where[1] != inf:
            where[1] = where[1] - lag
        other = other.shift(-lag)
    mask = self.isna() | other.isna()
    self = self.mask(mask)
    other = other.mask(mask)
    return (self * other).clip(*where).mean() - self.clip(*where).mean() * other.clip(
        *where
    ).mean()


@Appender(examples.corr_example, join="\n", indents=1)
def corr(self, other, where=(-inf, inf), lag=0, clip="pre"):
    """
    Calculates either correlation, autocorrelation or cross-correlation.

    All calculations are based off the `Pearson correlation coefficient <https://en.wikipedia.org/wiki/Pearson_correlation_coefficient>`_.

    The calculation is between two step functions described by *self* and *other*.
    If lag is None or 0 then correlation is calculated, otherwise cross-correlation is calculated.
    Autocorrelation is a special case of cross-correlation when *other* is equal to *self*.

    Parameters
    ----------
    other: :class:`Stairs`
        the stairs instance with which to compute the correlation
    where : tuple or list of length two, optional
        Indicates the domain interval over which to perform the calculation.
        Default is (-sc.inf, sc.inf) or equivalently (None, None).
    lag : int, float, pandas.Timedelta
        A pandas.Timedelta is only valid when domain is date-like.
    clip : {'pre', 'post'}, default 'pre'
        Only relevant when lag is non-zero.  Determines if the domain is applied before or
        after *other* is translated.  If 'pre' then the domain over which the calculation
        is performed is the overlap of the original domain and the translated domain.

    Returns
    -------
    float
        The correlation (or cross-correlation) between *self* and *other*

    See Also
    --------
    Stairs.cov, staircase.corr, staircase.cov
    """
    where = list(_replace_none_with_infs(where))
    if lag != 0:
        assert clip in ["pre", "post"]
        if clip == "pre" and where[1] != inf:
            where[1] = where[1] - lag
        other = other.shift(-lag)
    mask = self.isna() | other.isna()
    self = self.mask(mask)
    other = other.mask(mask)
    denominator = self.clip(*where).std() * other.clip(*where).std()
    if denominator == 0:
        return np.nan
    return self.cov(other, where) / denominator


def _get_stairs_method(name):
    return {
        "integral": integral,
        "mean": mean,
        "median": median,
        "mode": mode,
        "max": _max,
        "min": _min,
        "_max": _max,
        "_min": _min,
        "std": std,
        "var": var,
    }[name]



import pandas as pd

# registers the accessor with pandas
import staircase.core.arrays.accessor  # noqa
from staircase.constants import inf
from staircase.core.arrays import docstrings
from staircase.core.arrays.extension import StairsArray
from staircase.util._decorators import Appender


@Appender(docstrings.make_docstring("toplevel", "agg"), join="\n", indents=1)
def agg(collection, func):
    return StairsArray(collection).agg(func)


@Appender(docstrings.make_docstring("toplevel", "max"), join="\n", indents=1)
def max(collection):  # noqa
    return StairsArray(collection).max()


@Appender(docstrings.make_docstring("toplevel", "min"), join="\n", indents=1)
def min(collection):  # noqa
    return StairsArray(collection).min()


@Appender(docstrings.make_docstring("toplevel", "mean"), join="\n", indents=1)
def mean(collection):
    return StairsArray(collection).mean()


@Appender(docstrings.make_docstring("toplevel", "median"), join="\n", indents=1)
def median(collection):
    return StairsArray(collection).median()


@Appender(docstrings.make_docstring("toplevel", "sum"), join="\n", indents=1)
def sum(collection):  # noqa
    return StairsArray(collection).sum()


@Appender(docstrings.make_docstring("toplevel", "logical_or"), join="\n", indents=1)
def logical_or(collection):
    return StairsArray(collection).logical_or()


@Appender(docstrings.make_docstring("toplevel", "logical_and"), join="\n", indents=1)
def logical_and(collection):
    return StairsArray(collection).logical_and()


@Appender(docstrings.make_docstring("toplevel", "corr"), join="\n", indents=1)
def corr(collection, where=(-inf, inf)):
    return pd.Series(collection, dtype="Stairs").sc.corr(where)


@Appender(docstrings.make_docstring("toplevel", "cov"), join="\n", indents=1)
def cov(collection, where=(-inf, inf)):
    return pd.Series(collection, dtype="Stairs").sc.cov(where)


@Appender(docstrings.make_docstring("toplevel", "limit"), join="\n", indents=1)
def limit(collection, x, side="right"):
    return pd.Series(collection, dtype="Stairs").sc.limit(x, side)


@Appender(docstrings.make_docstring("toplevel", "sample"), join="\n", indents=1)
def sample(collection, x):
    return pd.Series(collection, dtype="Stairs").sc.sample(x)


@Appender(docstrings.make_docstring("toplevel", "plot"), join="\n", indents=1)
def plot(collection, ax=None, **kwargs):
    return pd.Series(collection, dtype="Stairs").sc.plot(ax, **kwargs)



import operator

import numpy as np
import pandas as pd

import staircase as sc
from staircase.core.ops import docstrings
from staircase.core.ops.common import _combine_stairs_via_values, requires_closed_match
from staircase.util import _sanitize_binary_operands
from staircase.util._decorators import Appender


def _make_relational_func(
    docstring, numpy_relational, series_relational, float_relational
):
    @Appender(docstring, join="\n", indents=1)
    @requires_closed_match
    def func(self, other):
        self, other = _sanitize_binary_operands(
            self, other
        )  # converts other to Stairs if not already
        if np.isnan(self.initial_value) or np.isnan(other.initial_value):
            initial_value = np.nan
        else:
            initial_value = (
                float_relational(self.initial_value, other.initial_value) * 1
            )

        if (
            (self._data is None and other._data is None)
            or (np.isnan(other.initial_value) and other._data is None)
            or (np.isnan(self.initial_value) and self._data is None)
        ):
            return sc.Stairs._new(
                initial_value=initial_value,
                data=None,
                closed=other.closed if np.isnan(self.initial_value) else self.closed,
            )
        elif self._data is None or other._data is None:
            if other._data is None:  # self._data exists
                values = self._get_values()
                new_values = numpy_relational(values, other.initial_value) * 1
                # new_values[values.isna()] = np.nan  # *1 converts bool to nan where applicable
                new_index = self._data.index
            else:  # other._data exists
                values = other._get_values()
                new_values = numpy_relational(self.initial_value, values)
                # new_values[values.isna()] = np.nan  # *1 converts bool to nan where applicable
                new_index = other._data.index

            new_instance = sc.Stairs._new(
                initial_value=initial_value,
                data=pd.DataFrame(
                    {"value": new_values},
                    index=new_index,
                ),
                closed=self.closed,
            )
            new_instance._remove_redundant_step_points()
            return new_instance
        else:
            return _combine_stairs_via_values(
                self, other, series_relational, float_relational
            )

    return func


def _is_series_equal(s1, s2):
    if not pd.api.types.is_float_dtype(s1):
        s1 = s1.astype("float64")
    if not pd.api.types.is_float_dtype(s2):
        s2 = s2.astype("float64")
    if pd.api.types.is_numeric_dtype(s1.index) and not pd.api.types.is_float_dtype(
        s1.index
    ):
        s1.index = s1.index.astype("float64")
    if pd.api.types.is_numeric_dtype(s2.index) and not pd.api.types.is_float_dtype(
        s2.index
    ):
        s2.index = s2.index.astype("float64")
    return pd.Series.equals(s1, s2)


def identical(self, other):
    """
    A boolean comparison for Stairs instances.

    Returns True if and only if *self* and *other* represent identical step functions.

    Parameters
    ----------
    other :  int, float, or :class:`Stairs`

    Returns
    -------
    boolean

    Examples
    --------

    >>> s1.identical(s1)
    True

    >>> s1.identical(s1.copy())
    True

    >>> s1.identical(s1.copy().layer(1,2))
    False

    >>> sc.Stairs(initial_value = 3).identical(3)
    True

    >>> sc.Stairs(initial_value = np.nan).identical(np.nan)
    True
    """
    self, other = _sanitize_binary_operands(self, other)
    if (self.initial_value != other.initial_value) and not (
        np.isnan(self.initial_value) and np.isnan(other.initial_value)
    ):
        return False
    elif self._data is None and other._data is None:
        return True
    elif self._data is None or other._data is None:
        return False
    elif self._valid_values and other._valid_values:
        return _is_series_equal(self._data["value"], other._data["value"])
    else:
        return _is_series_equal(self._get_deltas(), other._get_deltas())


lt = _make_relational_func(
    docstrings.lt_docstring,
    np.less,
    pd.Series.lt,
    operator.lt,
)


gt = _make_relational_func(
    docstrings.gt_docstring,
    np.greater,
    pd.Series.gt,
    operator.gt,
)


le = _make_relational_func(
    docstrings.le_docstring,
    np.less_equal,
    pd.Series.le,
    operator.le,
)


ge = _make_relational_func(
    docstrings.ge_docstring,
    np.greater_equal,
    pd.Series.ge,
    operator.ge,
)


eq = _make_relational_func(
    docstrings.eq_docstring,
    np.equal,
    pd.Series.eq,
    operator.eq,
)


ne = _make_relational_func(
    docstrings.ne_docstring,
    np.not_equal,
    pd.Series.ne,
    operator.ne,
)



import operator

import numpy as np
import pandas as pd

import staircase as sc
from staircase.core.ops import docstrings
from staircase.core.ops.common import _combine_stairs_via_values, requires_closed_match
from staircase.util import _sanitize_binary_operands
from staircase.util._decorators import Appender


def _make_boolean_func(docstring, series_comp, float_comp):
    @Appender(docstring, join="\n", indents=1)
    def func(self):
        if np.isnan(self.initial_value):
            initial_value = np.nan
        else:
            initial_value = float_comp(self.initial_value, 0) * 1

        if self._data is None:
            return sc.Stairs(initial_value=initial_value, closed=self.closed)
        values = series_comp(self._get_values(), 0) * 1
        values.loc[np.isnan(self._get_values().values)] = np.nan
        result = sc.Stairs._new(
            initial_value=initial_value,
            data=pd.DataFrame(
                {"value": values},
            ),
            closed=self.closed,
        )
        result._remove_redundant_step_points()
        return result

    return func


make_boolean = _make_boolean_func(
    docstrings.make_boolean_docstring, pd.Series.ne, operator.ne
)


invert = _make_boolean_func(docstrings.invert_docstring, pd.Series.eq, operator.eq)


def _make_logical_func(docstring, array_op, float_op):
    def _op_with_scalar_and(self, other):
        if np.isnan(other):
            return sc.Stairs._new(np.nan, None, closed=self.closed)
        elif other == 0:
            return sc.Stairs._new(0, None, closed=self.closed)
        else:
            return self.make_boolean()

    def _op_with_scalar_or(self, other):
        if np.isnan(other):
            return sc.Stairs._new(np.nan, None, closed=self.closed)
        elif other == 0:
            return self.make_boolean()
        else:
            return sc.Stairs._new(1, None, closed=self.closed)

    def _op_with_scalar_xor(self, other):
        if np.isnan(other):
            return sc.Stairs._new(np.nan, None, closed=self.closed)
        elif other == 0:
            return self.make_boolean()
        else:
            return self.invert()

    _op_with_scalar_func = {
        np.logical_and: _op_with_scalar_and,
        np.logical_or: _op_with_scalar_or,
        np.logical_xor: _op_with_scalar_xor,
    }[array_op]

    @Appender(docstring, join="\n", indents=1)
    @requires_closed_match
    def func(self, other):
        self, other = _sanitize_binary_operands(self, other)
        if other._data is None:
            return _op_with_scalar_func(self, other.initial_value)
        elif self._data is None:
            return _op_with_scalar_func(other, self.initial_value)
        else:
            return _combine_stairs_via_values(self, other, array_op, float_op)

    return func


def scalar_and(x, y):
    if np.isnan(x) or np.isnan(y):
        return np.nan
    if x == 0 or y == 0:
        return 0
    else:
        return 1


def scalar_or(x, y):
    if np.isnan(x) or np.isnan(y):
        return np.nan
    if x == 0 and y == 0:
        return 0
    else:
        return 1


def scalar_xor(x, y):
    if np.isnan(x) or np.isnan(y):
        return np.nan
    if x != 0 and y == 0:
        return 1
    elif x == 0 and y != 0:
        return 1
    else:
        return 0


logical_and = _make_logical_func(
    docstrings.logical_and_docstring, np.logical_and, scalar_and
)

logical_or = _make_logical_func(
    docstrings.logical_or_docstring, np.logical_or, scalar_or
)

logical_xor = _make_logical_func(
    docstrings.logical_xor_docstring, np.logical_xor, scalar_xor
)



sample_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> s3.plot(arrows=True)

>>> s3(1.5)
1

>>> s3(3)
nan

>>> s3([1, 3, 4])
array([ 1., nan, -1.])
"""

limit_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> s3.plot(arrows=True)

>>> s3.limit(2, side="left")
1.0

>>> s3.limit(2, side="right")
0.0

>>> s3.limit([2, 2.5, 3.5], side="right")
array([ 0., nan,  1.])
"""


layer_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> import staircase as sc
    ... (
    ...     sc.Stairs()
    ...     .layer(1,3)
    ...     .layer(4,5,-2)
    ...     .plot(arrows=True)
    ... )

.. plot::
    :context: close-figs

    >>> import pandas as pd
    >>> import staircase as sc
    >>> data = pd.DataFrame({"starts":[1,4,5.5],
    ...                      "ends":[3,5,7],
    ...                      "values":[-1,2,-3]})

    >>> data
    starts  ends  values
    0     1.0     3      -1
    1     4.0     5       2
    2     5.5     7      -3

    >>> (
    ...     sc.Stairs(initial_value=1.5)
    ...     .layer(data["starts"], data["ends"], data["values"])
    ...     .plot(arrows=True)
    ... )

.. plot::
    :context: close-figs

    >>> import pandas as pd
    >>> import staircase as sc
    >>> data = pd.DataFrame({"starts":[1,4,5.5],
    ...                      "ends":[3,5,7],
    ...                      "values":[-1,2,-3]})

    >>> data
    starts  ends  values
    0     1.0     3      -1
    1     4.0     5        2
    2     5.5     7     -3

    >>> (
    ...     sc.Stairs(initial_value=1.5)
    ...     .layer("starts", "ends", "values", frame=data)
    ...     .plot(arrows=True)
    ... )
"""

number_of_steps_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> s1.plot(arrows=True)
    >>> s1.number_of_steps
    5

.. plot::
    :context: close-figs

    >>> s3.plot(arrows=True)
    >>> s3.number_of_steps
    6
"""


shift_example = """
Examples
--------

.. plot::
    :context: close-figs
    :include-source: False

    >>> stair_list = [s2, s2.shift(1), s2.shift(-1)]
    >>> fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(8,3), sharey=True, sharex=True, tight_layout=True, dpi=400)
    >>> for ax, title, stair_instance in zip(axes, ("s2", "s2.shift(1)", "s2.shift(-1)"), stair_list):
    ...     stair_instance.plot(ax=ax, label=title, arrows=True)
    ...     ax.set_title(title)

Note that the definition of shift is designed to be consistent with :meth:`pandas.Series.shift`

>>> pd.Series(s2(range(7)))
0    0.5
1    0.5
2    0.0
3   -1.0
4   -1.0
5   -1.0
6    0.0
dtype: float64

>>> pd.Series(s2(range(7))).shift(1)
0    NaN
1    0.5
2    0.5
3    0.0
4   -1.0
5   -1.0
6   -1.0
dtype: float64

>>> pd.Series(s2.shift(1)(range(7)))
0    0.0
1    0.5
2    0.5
3    0.0
4   -1.0
5   -1.0
6   -1.0
dtype: float64
"""

diff_example = """
Examples
--------

.. plot::
    :context: close-figs
    :include-source: False

    >>> stair_list = [s2, s2.shift(1), s2.diff(1)]
    >>> fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(8,3), sharey=True, sharex=True, tight_layout=True, dpi=400)
    >>> for ax, title, stair_instance in zip(axes, ("s2", "s2.shift(1)", "s2.diff(1)"), stair_list):
    ...     stair_instance.plot(ax=ax, label=title, arrows=True)
    ...     ax.set_title(title)
    ... s2.plot(ax=axes[1], label="s2", linestyle="--", arrows=True)
    ... axes[1].legend()

Note that the definition of diff is designed to be consistent with :meth:`pandas.Series.diff`

>>> pd.Series(s2(range(7)))
0    0.5
1    0.5
2    0.0
3   -1.0
4   -1.0
5   -1.0
6    0.0
dtype: float64

>>> pd.Series(s2(range(7))).diff(1)
0    NaN
1    0.0
2   -0.5
3   -1.0
4    0.0
5    0.0
6    1.0
dtype: float64

>>> pd.Series(s2.diff(1)(range(7)))
0    0.5
1    0.0
2   -0.5
3   -1.0
4    0.0
5    0.0
6    1.0
dtype: float64
"""

describe_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> s2.plot(arrows=True)

>>> s2.describe()
unique    3.000000
mean     -0.272727
std       0.686349
min      -1.000000
25%      -1.000000
50%       0.000000
75%       0.500000
max       0.500000
dtype: float64

>>> s2.describe(where=(0,6), percentiles=np.linspace(0,100,6))
unique    3.000000
mean     -0.250000
std       0.661438
min      -1.000000
0%       -1.000000
20%      -1.000000
40%      -1.000000
60%       0.000000
80%       0.500000
100%      0.500000
max       0.500000
dtype: float64
"""

cov_example = """
Examples
--------

.. plot::
    :context: close-figs
    :include-source: False

    >>> fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(7,3), sharey=True, sharex=True, tight_layout=True, dpi=400)
    >>> for ax, title, stair_instance in zip(axes, ("s1", "s2"), (s1, s2)):
    ...     stair_instance.plot(ax=ax, label=title, arrows=True)
    ...     ax.set_title(title)

>>> s1.cov(s2)
0.1404958677685951

>>> s2.cov(s1)
0.1404958677685951

>>> s1.cov(s2, where=(0, 6))
0.125

>>> # autocovariance with lag 1
>>> s1.cov(s1, where=(1, 5), lag=1)
-0.3333333333333333

>>> # cross-covariance with lag 1
>>> s1.cov(s2, where=(1, 4.5), lag=1)
0.15999999999999998

>>> # cross-covariance with lag 1
>>> s1.cov(s2, where=(1, 4.5), lag=1, clip='post')
0.163265306122449
"""

corr_example = """
Examples
--------

.. plot::
    :context: close-figs
    :include-source: False

    >>> fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(7,3), sharey=True, sharex=True, tight_layout=True, dpi=400)
    >>> for ax, title, stair_instance in zip(axes, ("s1", "s2"), (s1, s2)):
    ...     stair_instance.plot(ax=ax, label=title, arrows=True)
    ...     ax.set_title(title)

>>> s1.corr(s2)
0.24687803791136045

>>> s2.corr(s1)
0.24687803791136045

>>> s1.corr(s2, where=(0, 6))
0.27500954910846337

>>> # autocorrelation with lag 1
>>> s1.corr(s1, where=(1, 5), lag=1)
-0.8660254037844386

>>> # cross-correlation with lag 1
>>> s1.corr(s2, where=(1, 5.5), lag=1)
0.4961389383568339

>>> # cross-correlation with lag 1
>>> s1.corr(s2, where=(1, 4.5), lag=1, clip='post')
0.4961389383568339
"""

rolling_mean_example = """
Examples
--------

>>> s2.rolling_mean(window=[-0.5, 0.5])
-0.5    0.0
 0.5    0.5
 1.5    0.5
 2.5    0.0
 3.5   -1.0
 5.0   -1.0
 6.0    0.0
dtype: float64

.. plot::
    :context: close-figs

    >>> series_list = [s2.rolling_mean(window=[-0.5, 0.5]), s2.rolling_mean(window=[-0.5, 0.5], where=(0, 5.5))]
    >>> fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(7,3), sharey=True, sharex=True, tight_layout=True, dpi=400)
    >>> for ax, title, series in zip(axes, ("default", "'where' specified"), series_list):
    ...     s2.plot(ax=ax)
    ...     series.plot(ax=ax, label='rolling mean')
    ...     ax.set_title(title)
    ...     ax.legend()

.. plot::
    :context: close-figs

    >>> series_list = [s2.rolling_mean(window=[-1, 0], where=(0, 5.5)), s2.rolling_mean(window=[0, 1], where=(0, 5.5))]
    >>> fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(8,3), sharey=True, sharex=True, tight_layout=True, dpi=400)
    >>> for ax, title, series in zip(axes, ("trailing window", "leading window"), series_list):
    ...     s2.plot(ax=ax)
    ...     series.plot(ax=ax, label='rolling mean')
    ...     ax.set_title(title)
    ...     ax.legend()
"""



from staircase.core.ops import docstrings
from staircase.util._decorators import Appender

from .arithmetic import add, divide, multiply, subtract
from .logical import logical_and, logical_or, logical_xor


@Appender(docstrings.radd_docstring, join="\n", indents=1)
def radd(self, other):
    return add(other, self)


@Appender(docstrings.rdivide_docstring, join="\n", indents=1)
def rdivide(self, other):
    return divide(other, self)


@Appender(docstrings.rmultiply_docstring, join="\n", indents=1)
def rmultiply(self, other):
    return multiply(other, self)


@Appender(docstrings.rsubtract_docstring, join="\n", indents=1)
def rsubtract(self, other):
    return subtract(other, self)


@Appender(docstrings.logical_rand_docstring, join="\n", indents=1)
def logical_rand(self, other):
    return logical_and(other, self)


@Appender(docstrings.logical_ror_docstring, join="\n", indents=1)
def logical_ror(self, other):
    return logical_or(other, self)


@Appender(docstrings.logical_rxor_docstring, join="\n", indents=1)
def logical_rxor(self, other):
    return logical_xor(other, self)
